import logging
from dataclasses import dataclass, replace

from typing_extensions import Self

from fuefpyco.ds import Failure, msum
from fuefpyco.pa import ComputationConfig, mapreduce


@dataclass
class AMonoid:
    a: int

    def __add__(self, other: Self) -> Self:
        return replace(self, a=self.a + other.a)

    @classmethod
    def empty(cls) -> Self:
        return cls(0)


_error = ValueError("thou shalt not pass more than 9")


def simple_f(a: int):
    if a > 9:
        raise _error
    return AMonoid(a=a * 2)


def test_mapreduce_happy():
    logging.basicConfig(level="DEBUG", force=True)
    config = ComputationConfig(parallelism=4, single_task_timeout_s=1)
    input_seq = [1, 2, 3]
    result = mapreduce(simple_f, input_seq, config)
    assert result.result == msum((simple_f(e) for e in input_seq), AMonoid)
    assert result.failure == []

    seq_with_error = [1, 2, 10]
    result = mapreduce(simple_f, seq_with_error, config)
    assert result.result == msum((simple_f(e) for e in seq_with_error[:2]), AMonoid)
    assert result.failure == [Failure("failure with args 10", _error)]



"""
Module contents:
 - Monoid: a Protocol representing "things you can sum together". Useful for running concurrent computation and then
   merging results together. Accompandied with `msum` function.
 - MaybeResult and Failure: useful for computations where you don't want the first exception to crash everything down,
   but rather finish what can be finished and collect all exceptions at the end.
"""

from abc import abstractmethod
from dataclasses import dataclass, replace
from typing import Any, Generic, Iterable, Optional, Protocol, Type, TypeVar, runtime_checkable

from typing_extensions import Self


# *** Monoid ***
@runtime_checkable
class Monoid(Protocol):
    """This simplifies applying map-reduce style processing logic. Imagine you have a list of urls of csvs you want
    to download and concat as a pandas. Or a list of dictionaries, each of which has some field 'k' you want to sum
    together. You can do that with for cycles, comprehensions, functools.reduce, ... Or, you can just define
    a dataclass that represents the result of a single computation, and how two computations can be put together
    (for the examples above, `pd.concat`, `+`), to satisfy the Monoid protocol, and then the `msum` function does
    the rest for you."""

    @abstractmethod
    def __add__(self, other: Self) -> Self:
        raise NotImplementedError

    @classmethod
    @abstractmethod
    def empty(cls) -> Self:
        raise NotImplementedError


# NOTE it is rather unfortunate that the following method needs the Type explicitly -- because Python has, afaik,
# no proper template/macro capability to infer that information from the context. Also, we can't just turn it
# into a classmethod, if we want Monoid to stay a protocol and not become an ABC.
# NOTE going with Optional would not be much better, because the add/radd result would have wrong type
TMonoid = TypeVar("TMonoid", bound=Monoid)


def msum(i: Iterable[TMonoid], t: Type[TMonoid]) -> TMonoid:
    return sum(i, start=t.empty())


@dataclass
class Failure:
    """Represents a caught Exception. User fills `origin` based on context -- it can be stack trace, params,
    identifier, ... To be used in concurrent compute, when single Exception should not bring it all down."""

    origin: str
    exception: Exception

    def __eq__(self, other: Any) -> bool:
        # NOTE we override since `Exception`'s eq seems to be non cooperative with pickling
        if not isinstance(other, Failure):
            return False
        if other.origin == self.origin and str(self.exception) == str(other.exception):
            return True
        return False


@dataclass
class MaybeResult(Generic[TMonoid]):
    """Note this is *not* an Either-class, both `result` and `failure` may be filled. This is to support non-critical
    failures, and previous-attempt failures"""

    result: Optional[TMonoid]  # ideally, this would be just TMonoid. Alas, because of type erasure we couldnt `empty()`
    failure: list[Failure]

    @classmethod
    def empty(cls) -> Self:
        return cls(result=None, failure=[])

    def __add__(self, other: Self) -> Self:
        # annoyance due to python Optional not being well combinable. TODO improve
        if self.result is None and other.result is None:
            result = None
        elif self.result is None:
            result = other.result
        elif other.result is None:
            result = self.result
        else:
            result = self.result + other.result
        return replace(self, result=result, failure=self.failure + other.failure)



from dataclasses import dataclass, replace

from typing_extensions import Self

from fuefpyco.ds import Failure, MaybeResult, msum


@dataclass
class AMonoid:
    a: int

    def __add__(self, other: Self) -> Self:
        return replace(self, a=self.a + other.a)

    @classmethod
    def empty(cls) -> Self:
        return cls(0)


def test_monoid() -> None:
    l1 = [AMonoid(4), AMonoid(5)]
    l2: list[AMonoid] = []
    assert msum(l1, AMonoid) == AMonoid(a=9)
    assert msum(l2, AMonoid) == AMonoid(a=0)


def test_maybe_result() -> None:
    succ1 = MaybeResult(result=AMonoid(a=4), failure=[])
    fail1 = MaybeResult(result=AMonoid.empty(), failure=[Failure(origin="a", exception=ValueError())])
    succ2 = MaybeResult(result=AMonoid(a=5), failure=[])
    fail2 = MaybeResult(result=AMonoid.empty(), failure=[Failure(origin="b", exception=ValueError())])
    all_r = msum([succ1, fail1, succ2, fail2], MaybeResult)
    assert all_r.result == AMonoid(a=9)
    assert all_r.failure == [fail1.failure[0], fail2.failure[0]]



"""
Module contents:
    - flatmap -- standard functional construct, useful for e.g. map-and-filter or maps that produce Optional results,
    - fold_transform -- simpler application of a pipeline of transformations on a single object,
    - unzip -- just a name for the inverse zip operation,
    - consume -- return first n elements of iterator, and then the rest of the iterator. Generalisation of head-tail,
    - windows -- `windows([1,2,3,4,5], 2) -> [[1, 2], [3,4], [5]].`

This whole module is based on iterators/generators, no unneeded list allocations are happening

"""
from functools import reduce
from itertools import chain
from typing import Callable, Iterable, Iterator, Optional, TypeVar

TA = TypeVar("TA")
TB = TypeVar("TB")


def flatmap(f: Callable[[TA], Iterable[TB]], xs: Iterable[TA]) -> Iterable[TB]:
    """Often one wants to map-and-filter a sequence, which perfectly suits the flatMap."""
    return (y for x in xs for y in f(x))


def fold_transform(obj: TA, pipeline: Iterable[Callable[[TA], TA]]) -> TA:
    """Say you have a dataframe and want to apply a chain of transformations on it"""
    return reduce(lambda obj, func: func(obj), pipeline, obj)


def unzip(zipped: Iterable[tuple[TA, TB]]) -> tuple[Iterable[TA], Iterable[TB]]:
    """Inverse operation to zip: [(a, b), (c, d), (e, f)] => ((a, c, e), (b, d, f))"""
    # NOTE this is basically just a type-anotated more-readable alias
    a, b = zip(*zipped)
    return a, b


def consume(it: Iterator[TA], n: Optional[int]) -> tuple[list[TA], Iterator[TA]]:
    """Retuns up to first `n` elements of an iterator, and the remaining iterator part. Generalises head-tail."""
    # TODO return iterator instead of list?
    ls = []
    i = 0
    if n is None:
        return list(it), iter(())
    while i < n:
        try:
            ls.append(next(it))
            i += 1
        except StopIteration:
            break
    return ls, it


def windows(s: Iterable[TA], n: int) -> Iterable[Iterable[TA]]:
    """windows([1,2,3,4,5], 2) -> [[1, 2], [3,4], [5]].
    Beware that this shares a similar quirk as groupby -- don't advance the outer iterator until the inner iterator
    is wholly consumed. In other words:
    ```
    for window in windows(s, n):
        for element in window:
            work(element)
    ```
    is fine, but list(windows(s, n)) is not because it does not consume individual elements."""
    it = iter(s)

    def _take_n(it: Iterator[TA], n: int) -> Iterable[TA]:
        while n > 0:
            try:
                yield next(it)
            except StopIteration:
                break
            n = n - 1

    while True:
        try:
            head_w = _take_n(it, n).__iter__()
            head_w_head = next(head_w)
            yield chain([head_w_head], head_w)
        except StopIteration:
            break



"""
This module provides parallel execution to map-reduce computations: single function to be applied on a sequence
of arguments, and the result being a monoid sum. The primary advantages over existing python high level interfaces are:
 - timeouts on the process level -- even when the GIL is held by the computation, you can still time it out
 - argument check by mypy. We enforce the function to be executed takes a single argument, typically a dataclass
   of the actual parameters. This restriction is due to **kwargs being quite hard to check.
 - MaybeResult-based exception propagation and process-level retries -- we don't want a single crash to bring the whole
   computation down. Note that if the function itself already returns MaybeResult, those get nested.

Notes:
  - Retry applies in the case of timeout or any exception thrown by the function. A case of non-zero exit of
    the process with _any_ result returned is considered a success, and the non-zero exit code is provided
    in a message in the `failure.origin` field.
  - In case of a retry, the failure reason(s) are returned as well, with the attempt counter prepended to
    the `failure.origin` field.
  - We use mp.Queue to retrieve the results from local workers. In case of huge data volumes, this may bring some
    troubles -- you may want to write disks instead, with the ipc being the file names only.
"""

import logging
import time
from dataclasses import dataclass
from multiprocessing import Process, Queue, get_context
from queue import Empty as EmptyQueueException
from typing import Callable, Generic, Iterable, Iterator, Literal, Optional, TypeVar

from fuefpyco.ds import Failure, MaybeResult, TMonoid

logger = logging.getLogger(__name__)

_process_join_grace = 3  # number of seconds we wait to join processes that returned. 0 should suffice


@dataclass
class ComputationConfig:
    parallelism: int
    single_task_timeout_s: int
    retry: int = 0
    implementation: Literal["process"] = "process"  # TODO add processes with reusal


T = TypeVar("T")


@dataclass
class _IntermediateResult(Generic[TMonoid]):
    result: MaybeResult[TMonoid]
    p_id: int


@dataclass
class _ProgressTracker(Generic[T]):
    p: Process
    arg: T
    submit_time_s: int


def _entrypoint(f: Callable[[T], TMonoid], arg: T, q: Queue, p_id: int):
    try:
        result = MaybeResult(f(arg), [])
    except Exception as e:
        result = MaybeResult(None, [Failure(f"failure with args {arg}", e)])
    q.put(_IntermediateResult(result=result, p_id=p_id))


def mapreduce(f: Callable[[T], TMonoid], s: Iterable[T], c: ComputationConfig) -> MaybeResult[TMonoid]:
    if c.retry > 0:
        raise NotImplementedError("non-zero retries are not implemented yet")
    result: MaybeResult[TMonoid] = MaybeResult.empty()
    ctx = get_context("forkserver")
    queue = ctx.Queue()

    running: dict[int, _ProgressTracker] = {}
    capacity: Callable[[], int] = lambda: c.parallelism - len(running)
    it: Optional[Iterator[T]] = iter(s)
    p_id = 0

    while True:
        # maybe submit new process
        if capacity() > 0 and it is not None:
            try:
                arg = next(it)
            except StopIteration:
                it = None
                continue
            p = Process(target=_entrypoint, args=(f, arg, queue, p_id))
            tracker = _ProgressTracker(p, arg, int(time.monotonic_ns() // 1e9))
            running[p_id] = tracker
            p.start()
            logging.debug(f"submitted a process #{p_id} with pid p.pid")
            p_id += 1
            continue
        # maybe quit
        if capacity() == c.parallelism:
            if running:
                raise ValueError(f"internal error: expected no processes running, but {len(running)} entries remain")
            logger.debug("all processes completed, breaking main loop")
            break
        # wait for next result
        oldest_submit = min(e.submit_time_s for e in running.values())
        next_timeout = oldest_submit + c.single_task_timeout_s
        current_time = time.monotonic_ns() // 1e9
        if next_timeout > current_time:
            try:
                wait_time = next_timeout - current_time
                logger.debug(f"about to wait for {wait_time} seconds for a result")
                intermediate = queue.get(timeout=wait_time)
                tracker = running.pop(intermediate.p_id)
                tracker.p.join(_process_join_grace)
                exitcode = tracker.p.exitcode
                if exitcode is None:
                    raise ValueError(f"process {tracker.p.pid} with arg {tracker.arg} failed to terminate")
                if exitcode != 0:
                    nonzero: MaybeResult[TMonoid] = MaybeResult(
                        None, [Failure(f"non-zero exit code {exitcode} with arg {tracker.arg}", Exception())]
                    )
                    intermediate.result = intermediate.result + nonzero
                result = result + intermediate.result
                logger.debug(f"a result for {intermediate.p_id} processed")
                continue
            except EmptyQueueException:
                logger.debug("wait unsuccessful")
                pass
        # kill
        for p_id in running:
            logger.debug(f"inspecting process #{p_id}")
            if running[p_id].submit_time_s < next_timeout:
                convict = running.pop(p_id)
                sentence: MaybeResult[TMonoid] = MaybeResult(
                    None, [Failure(f"timed out with arg {convict.arg}", Exception())]
                )
                result = result + sentence
                logger.debug(f"killing process #{p_id} with pid {convict.p.pid}")
                convict.p.kill()
                break
    return result



import multiprocessing
import os

from setuptools import setup

if __name__ == "__main__":
    # Freeze to support parallel compilation when using spawn instead of fork
    multiprocessing.freeze_support()
    setup(version=os.getenv("TARGET_VERSION"))



from fuefpyco.it import consume, flatmap, fold_transform, unzip, windows


def test_flatmap() -> None:
    c = [1, 3, 2, 4]
    r = flatmap(lambda v: [v] if v > 2 else [], c)
    assert list(r) == [3, 4]


def test_fold_transform() -> None:
    pipeline = [
        lambda a: a + 1,
        lambda a: a * 2,
    ]
    assert fold_transform(2, pipeline) == 6


def test_unzip() -> None:
    l1 = [1, 2, 3]
    l2 = ["a", "b", "c"]
    zl = zip(l1, l2)
    rl1, rl2 = unzip(zl)
    assert (list(rl1), list(rl2)) == (l1, l2)


def test_consume() -> None:
    ls = [1, 2, 3, 4, 5]
    li = iter(ls)
    nil, li = consume(li, 0)
    assert nil == []
    head, li = consume(li, 1)
    assert head[0] == 1
    body, li = consume(li, 2)
    assert body == [2, 3]
    assert list(li) == [4, 5]


def test_windows() -> None:
    ls = [1, 2, 3, 4, 5]
    badly_consumed = list(windows(ls, 2))
    badly_consumed_eval = [list(e) for e in badly_consumed]
    assert badly_consumed_eval == [[1], [2], [3], [4], [5]]
    well_consumed = []
    for window in windows(ls, 2):
        well_consumed.append(list(window))
    assert well_consumed[0] == [1, 2]
    assert well_consumed[1] == [3, 4]
    assert well_consumed[2] == [5]
